unit DDUtils;

interface

uses
   Windows, DDVer, DDraw;

var
   ResourceDLL : integer = 0;

// DirectDrawSurface  Ʈ ׸
procedure DDCopyBitmap(Surface             : II_DirectDrawSurface;
                       Bitmap              : HBITMAP;
                       x, y, Width, Height : integer);

// ϴ  ǽ 
function  DDCreateSurface(DirectDraw       : II_DirectDraw;
                          Width            : integer;
                          Height           : integer;
                          UseSystemMemory  : boolean) : II_DirectDrawSurface;

// Ʈ ҽ Ǵ Ϸ DirectDrawSurface  Ѵ
function  DDLoadBitmap(DirectDraw       : II_DirectDraw;
                       const BitmapName : string;
                       Width            : integer;
                       Height           : integer;
                       UseSystemMemory  : boolean) : II_DirectDrawSurface;

// directdraw surface   Ǵ ҽ  Ʈ εѴ
//  restore  Ŀ surface  ٽ εҶ ȴ.
procedure DDReLoadBitmap(Surface          : II_DirectDrawSurface;
                         const BitmapName : string);

// PCX ׸ Surface о δ.
procedure DDReLoadPCX(Surface       : II_DirectDrawSurface;
                      const PCXName : string);

// Ʈ ҽ Ǵ Ϸ DirectDraw palette  Ѵ
//  ҽ  ʰų '' ̸, Ʈ 332 ȷƮ 
function  DDLoadPalette(DirectDraw       : II_DirectDraw;
                        const BitmapName : string) : IDirectDrawPalette;

// RGB ÷  ÷ ٲ۴
// GDI Լ SetPixel()   յǴ ÷ ã 
// ޸𸮸 Lock ϸ Ѵ
function  DDColorMatch(Surface : II_DirectDrawSurface;
                       RGB     : TColorRef) : integer;

// Surface   ־ RGB  ̿Ͽ  Ѵ
//   CLR_INVALID  Ѿ´ٸ    üѴ
function  DDSetColorKey(Surface : II_DirectDrawSurface;
                        RGB     : TColorRef) : HResult;

// Surface    
procedure DDClearDevice(Surface : II_DirectDrawSurface; Color : Word);

implementation

procedure DDCopyBitmap(Surface             : II_DirectDrawSurface;
                       Bitmap              : HBITMAP;
                       x, y, Width, Height : integer);
var
   ImageDC     : HDC;
   DC          : HDC;
   BM          : Windows.TBitmap;
   SurfaceDesc : TT_DDSurfaceDesc;
begin

   if (Surface = NIL) or (Bitmap = 0) then
    {$IFDEF _DEBUG}
      Raise Exception.Create('Invalid parameters for DDCopyBitmap');
    {$ELSE}
      exit;
    {$ENDIF}

   // make sure this surface is restored.
   Surface._Restore;

   // select bitmap into a memoryDC so we can use it.
   ImageDC := CreateCompatibleDC(0);

   try
      SelectObject(ImageDC, Bitmap);

      // get size of the bitmap
      GetObject(Bitmap, SizeOf(BM), @BM);
      if Width  = 0 then Width  := BM.bmWidth;
      if Height = 0 then Height := BM.bmHeight;

      // get size of surface.
      SurfaceDesc.dwSize  := SizeOf(SurfaceDesc);
      SurfaceDesc.dwFlags := DDSD_HEIGHT or DDSD_WIDTH;
      Surface.GetSurfaceDesc(SurfaceDesc);

      if Surface.GetDC(DC) <> DD_OK then
       {$IFDEF _DEBUG}
         Raise Exception.Create('GetDC failed for DirectDraw surface');
       {$ELSE}
         exit;
       {$ENDIF}

      try
         StretchBlt(DC,0,0,SurfaceDesc.dwWidth,SurfaceDesc.dwHeight,
                    ImageDC,x,y,Width,Height,SRCCOPY);
      finally
         Surface.ReleaseDC(DC);
      end;

   finally
     DeleteDC(ImageDC);
   end;
end;

function  DDCreateSurface(DirectDraw       : II_DirectDraw;
                          Width            : integer;
                          Height           : integer;
                          UseSystemMemory  : boolean) : II_DirectDrawSurface;
var
   SurfaceDesc : TT_DDSurfaceDesc;
   ColorKey    : TDDColorKey;
begin

   // create a DirectDrawSurface for this bitmap
   FillChar(SurfaceDesc, SizeOf(SurfaceDesc), 0);
   with SurfaceDesc do begin
      dwSize         := SizeOf(SurfaceDesc);
      dwFlags        := DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH;
      ddsCaps.dwCaps := DDSCAPS_OFFSCREENPLAIN;
      if UseSystemMemory then
         ddSCaps.dwCaps := ddSCaps.dwCaps or DDSCAPS_SYSTEMMEMORY;
      dwWidth        := Width;
      dwHeight       := Height;
   end;

   if DirectDraw.CreateSurface(SurfaceDesc,Result,NIL) <> DD_OK then
   {$IFDEF _DEBUG}
     Raise Exception.Create('CreateSurface failed');
   {$ELSE}
     exit;
   {$ENDIF}

   if Assigned(Result) then begin
      ColorKey.dwColorSpaceLowValue  := 0;
      ColorKey.dwColorSpaceHighValue := 0;
      Result.SetColorKey(DDCKEY_SRCBLT,@ColorKey);
   end;

end;

function  DDLoadBitmap(DirectDraw       : II_DirectDraw;
                       const BitmapName : string;
                       Width            : integer;
                       Height           : integer;
                       UseSystemMemory  : boolean) : II_DirectDrawSurface;
var
   Bitmap      : HBitmap;
   BM          : Windows.TBitmap;
   SurfaceDesc : TT_DDSurfaceDesc;
begin
   Bitmap := 0;
   if BitmapName <> '' then begin
      //  try to load the bitmap as a resource, if that fails, try it as a file
      if ResourceDLL <> 0 then begin
         Bitmap := LoadImage(ResourceDLL,PChar(BitmapName),
                             IMAGE_BITMAP,Width,Height,LR_CREATEDIBSECTION);
      end else begin
         Bitmap := LoadImage(GetModuleHandle(NIL),PChar(BitmapName),
                             IMAGE_BITMAP,Width,Height,LR_CREATEDIBSECTION);
      end;
   end;

   try
      if BitmapName <> '' then begin
         if Bitmap = 0 then
            Bitmap := LoadImage(0,PChar(BitmapName+'.BMP'),IMAGE_BITMAP,Width,Height,
                                LR_LOADFROMFILE or LR_CREATEDIBSECTION);
            if Bitmap = 0 then
               Bitmap := LoadImage(0,PChar(BitmapName+'.DAT'),IMAGE_BITMAP,Width,Height,
                                   LR_LOADFROMFILE or LR_CREATEDIBSECTION);
         if Bitmap = 0 then
          {$IFDEF _DEBUG}
            Raise Exception.CreateFmt('Unable to load bitmap %s', [ BitmapName ]);
          {$ELSE}
            exit;
          {$ENDIF}

         // get size of the bitmap
         GetObject(Bitmap, SizeOf(BM), @BM);

      end else begin
         BM.bmWidth  := Width;
         BM.bmHeight := Height;
      end;

      // create a DirectDrawSurface for this bitmap
      FillChar(SurfaceDesc, SizeOf(SurfaceDesc), 0);
      with SurfaceDesc do begin
         dwSize         := SizeOf(SurfaceDesc);
         dwFlags        := DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH;
         ddsCaps.dwCaps := DDSCAPS_OFFSCREENPLAIN;
         if UseSystemMemory then
            ddSCaps.dwCaps := ddSCaps.dwCaps or DDSCAPS_SYSTEMMEMORY;
         dwWidth        := BM.bmWidth;
         dwHeight       := BM.bmHeight;
      end;

      if DirectDraw.CreateSurface(SurfaceDesc,Result,NIL) <> DD_OK then
      {$IFDEF _DEBUG}
        Raise Exception.Create('CreateSurface failed');
      {$ELSE}
        exit;
      {$ENDIF}

      if BitmapName <> '' then begin
         DDCopyBitmap(Result,Bitmap,0,0,0,0);
      end;   

   finally
      if Bitmap      <> 0 then DeleteObject(Bitmap);
   end;
end;

procedure DDReLoadBitmap(Surface          : II_DirectDrawSurface;
                         const BitmapName : string);

var
   Bitmap : HBitmap;

begin
   //  try to load the bitmap as a resource, if that fails, try it as a file
   if ResourceDLL <> 0 then begin
      Bitmap := LoadImage(ResourceDLL,PChar(BitmapName),
                          IMAGE_BITMAP,0,0,LR_CREATEDIBSECTION);
   end else begin
      Bitmap := LoadImage(GetModuleHandle(NIL),PChar(BitmapName),
                          IMAGE_BITMAP,0,0,LR_CREATEDIBSECTION);
   end;

   try
      if Bitmap = 0 then
         Bitmap := LoadImage(0,PChar(BitmapName+'.BMP'),IMAGE_BITMAP,
                             0,0,LR_LOADFROMFILE or LR_CREATEDIBSECTION);
      if Bitmap = 0 then
       {$IFDEF _DEBUG}
         Raise Exception.CreateFmt('Unable to load bitmap %s', [ BitmapName ]);
       {$ELSE}
         exit;
       {$ENDIF}

      DDCopyBitmap(Surface, Bitmap, 0, 0, 0, 0);

   finally
      DeleteObject(Bitmap);
   end;
end;

procedure LoadPCX256(file_name : string; lpSurface : Pointer; lPitch : integer; dx, dy : word; Is15Bit : Boolean);
const
   MAX_PALETTE = 256;
   MAX_X_LINE  = 640;
   MAX_Y_LINE  = 480;
type
   TPaletteData = array[0..pred(MAX_PALETTE)] of record
      pdRED, pdGREEN, pdBLUE : byte;
   end;
var
   f             : file;
   buffer        : array[0..2047] of byte;
   actual_read   : integer;
   position, count, wide, height, end_line, pack : word;
   i, x, y       : word;
   palette_data  : TPaletteData;
   palette_ref   : array[0..Pred(MAX_PALETTE)] of word;
   m_buffer      : array[0..Pred(MAX_Y_LINE),0..Pred(MAX_X_LINE)] of byte;
   pw            : PWord;
begin
   if Pos('.',file_name) = 0 then file_name := file_name + '.PCX';

   AssignFile(f,file_name);
  {$I-}
   Reset(f, 1);
  {$I+}
   if IOResult <> 0 then begin
      MessageBeep(0);
      exit;
   end;

   Seek(f,FileSize(f) - 3 shl 8 - 1);
   BlockRead(f,buffer,3 shl 8 + 1);
   if buffer[0] = 12 then begin
      move(buffer[1],palette_data,MAX_PALETTE*3);
      for i := 0 to pred(MAX_PALETTE) do begin
         with palette_data[i] do
         if Is15Bit then
            palette_ref[i] := (Word(pdRED shr 3) shl 10) or (Word(pdGREEN shr 3) shl 5) or Word(pdBLUE shr 3)
         else
            palette_ref[i] := (Word(pdRED shr 3) shl 11) or (Word(pdGREEN shr 2) shl 5) or Word(pdBLUE shr 3)
//         palette_ref[i] := RGB(pdRED,pdGREEN,pdBLUE);
      end;
      for i := 1 to 3 shl 8 + 1 do
         buffer[i] := buffer[i] shr 2;
      move(buffer[1],palette_data,MAX_PALETTE*3);
   end;

   x := 0; y := 0;
   seek(f,0);
   BlockRead(f,buffer,128,actual_read);
   if (buffer[0] <> 10) or (buffer[3] <> 8) then begin
      write(^G);
      closeFile(f);
      exit;
   end;

   wide   := succ((buffer[9] - buffer[5]) shl 8 + buffer[8] - buffer[4]);
   height := succ((buffer[11] - buffer[7]) shl 8 + buffer[10] - buffer[6]);
   pack   := 0; count := 0; end_line := y + height;
   repeat
      BlockRead(f,buffer,2048,actual_read);
      position := 0;
      while (position < actual_read) and (y < end_line) do begin
         if pack <> 0 then begin
            fillChar(m_buffer[y+dy,x+count+dx],pack,buffer[position]);
            count := count + pack;
            pack := 0;
         end else if (buffer[position] and $C0) = $C0 then
            pack := buffer[position] and $3F
         else begin
            m_buffer[y+dy,x+count+dx] := buffer[position];
            inc(count);
         end;
         inc(position);
         if count = wide then begin
            count := 0;
            inc(y);
         end;
      end;
   until (actual_read = 0) or (y = end_line);

   closeFile(f);

   for y := 0 to pred(MAX_Y_LINE) do begin
      pw := PWord(DWORD(lpSurface) + y * lPitch);
      for x := 0 to pred(MAX_X_LINE) do begin
         pw^ := palette_ref[m_buffer[y,x]];
         Inc(pw);
      end;
   end;

end;

procedure DDReLoadPCX(Surface       : II_DirectDrawSurface;
                      const PCXName : string);
var
   DDSurfaceDesc : TT_DDSurfaceDesc;
begin
   DDSurfaceDesc.dwSize  := sizeof(DDSurfaceDesc);
   Surface.Lock(PRect(nil),DDSurfaceDesc,DDLOCK_SURFACEMEMORYPTR+DDLOCK_WAIT,0);
   LoadPCX256(PCXName,DDSurfaceDesc.lpSurface,DDSurfaceDesc.lPitch,0,0, TRUE);
   Surface.UnLock(DDSurfaceDesc.lpSurface);
end;

function  DDLoadPalette(DirectDraw       : II_DirectDraw;
                        const BitmapName : string) : IDirectDrawPalette;

type
   TRGB = array[0..255] of TRGBQuad;
   PRGB = ^TRGB;
var
   i, n        : integer;
   h           : HRsrc;
   BitmapInfo  : PBitmapInfo;
   APE         : array[ 0..255 ] of TPaletteEntry;
   RGB         : PRGB;
   bfHeader    : TBitmapFileHeader;
   biHeader    : TBitmapInfoHeader;
   Temp        : byte;
   fp          : File;
begin
   // build a 332 palette as the default.
   for i := 0 to 255 do with APE[ i ] do begin
      peRed   := (((i SHR   5) and $07) * 255 div 7);
      peGreen := (((i SHR   2) and $07) * 255 div 7);
      peBlue  := ( (i and $03)          * 255 div 3);
      peFlags := 0;
   end;

   // get a pointer to the bitmap resource.
   if BitmapName <> '' then begin
      h := FindResource(0,PChar(BitmapName),RT_BITMAP);
      if h <> 0 then begin
         BitmapInfo := PBitmapInfo(LockResource(LoadResource(0,h)));
         RGB        := PRGB(@BitmapInfo^.bmiColors);
         if (BitmapInfo = NIL) or
            (BitmapInfo^.bmiHeader.biSize < sizeof(TBITMAPINFOHEADER)) then n := 0
         else
         if (BitmapInfo^.bmiHeader.biBitCount > 8) then n := 0
         else
         if (BitmapInfo^.bmiHeader.biClrUsed = 0) then
            n := 1 SHL BitmapInfo^.bmiHeader.biBitCount
         else
            n := BitmapInfo^.bmiHeader.biClrUsed;

         //  a DIB color table has its colors stored BGR not RGB
         //  so flip them around.
         for i := 0 to n - 1 do with APE[ i ], RGB^[ i ] do begin
            peRed   := rgbRed;
            peGreen := rgbGreen;
            peBlue  := rgbBlue;
            peFlags := 0;
         end;

      end else begin

         AssignFile(fp, BitmapName + '.BMP');
       {$I-}
         Reset(fp, 1);
       {$I+}
         if IOResult = 0 then begin
            BlockRead(fp, bfHeader, SizeOf(bfHeader));
            BlockRead(fp, biHeader, SizeOf(biHeader));
            BlockRead(fp, APE, SizeOf(APE));

            CloseFile(fp);
         end;

         //  get the number of colors in the color table
         if biHeader.biSize <> SizeOf(TBitmapInfoHeader) then n := 0
         else
         if biHeader.biBitCount > 8 then n := 0
         else
         if biHeader.biClrUsed = 0 then n := 1 SHL biHeader.biBitCount
         else
            n := biHeader.biClrUsed;

         //  a DIB color table has its colors stored BGR not RGB
         //  so flip them around.
         for i := 0 to n - 1 do with APE[ i ] do begin
            Temp   := peRed;
            peRed  := peBlue;
            peBlue := Temp;
         end;
      end;
   end;

   // create the DD palette
   if DirectDraw.CreatePalette(DDPCAPS_8BIT, @APE[ 0 ], Result, NIL) <> DD_OK then
   {$IFDEF _DEBUG}
     Raise Exception.Create('DirectDraw.CreatePalette failed');
   {$ELSE}
     exit;
   {$ENDIF}
end;

function  DDColorMatch(Surface : II_DirectDrawSurface;
                       RGB     : TColorRef) : integer;
var
   TempValue   : TColorRef;
   DC          : HDC;
   SurfaceDesc : TT_DDSurfaceDesc;
   ARect       : TRect;
   DDResult    : HResult;
begin
   Result    := CLR_INVALID;

   TempValue := 0;
   //  use GDI SetPixel to color match for us
   if (RGB <> CLR_INVALID) and (Surface.GetDC(DC) = DD_OK) then
   try
     TempValue := GetPixel(DC,0,0);     // save current pixel value
     SetPixel(DC,0,0,RGB);              // set our value
   finally
     Surface.ReleaseDC(DC);
   end;

   // now lock the surface so we can read back the converted color
   SurfaceDesc.dwSize := sizeof(SurfaceDesc);
   try
      repeat
         FillChar(ARect, sizeof(ARect), 0);
         DDResult := Surface.Lock(@ARect,SurfaceDesc,0,0);
      until DDResult <> DDERR_WASSTILLDRAWING;

      if DDResult = DD_OK then begin
        // get value and mask it to bpp
         Result := PInteger(SurfaceDesc.lpSurface)^ and
                            (1 SHL SurfaceDesc.ddpfPixelFormat.dwRGBBitCount) - 1;
      end;
   finally
      Surface.Unlock(NIL);
   end;

   //  now put the color that was there back.
   if (RGB <> CLR_INVALID) and (Surface.GetDC(DC) = DD_OK) then
   try
     SetPixel(DC,0,0,TempValue);
   finally
     Surface.ReleaseDC(DC);
   end;
end;

function  DDSetColorKey(Surface : II_DirectDrawSurface;
                        RGB     : TColorRef) : HResult;
var
   ColorKey : TDDColorKey;
begin
   ColorKey.dwColorSpaceLowValue  := DDColorMatch(Surface,RGB);
   ColorKey.dwColorSpaceHighValue := ColorKey.dwColorSpaceLowValue;
   Result                         := Surface.SetColorKey(DDCKEY_SRCBLT,@ColorKey);
end;

procedure DDClearDevice(Surface : II_DirectDrawSurface; Color : Word);
var
   DDBltFX       : TDDBltFX;
   DDSurfaceDesc : TT_DDSurfaceDesc;
begin
   FillChar(DDBltFX,sizeof(DDBltFX),0);
   DDBltFX.dwSize      := sizeof(TDDBltFX);
   DDBltFX.dwFillColor := Color;
   if Surface.Blt(PRect(nil),nil,PRect(nil),DDBLT_COLORFILL,@DDBltFX) <> DD_OK then begin
      FillChar(DDSurfaceDesc,sizeof(DDSurfaceDesc),0);
      DDSurfaceDesc.dwSize  := sizeof(DDSurfaceDesc);
      DDSurfaceDesc.dwFlags := DDSD_ALL;
      Surface.Lock(PRect(nil),DDSurfaceDesc,DDLOCK_SURFACEMEMORYPTR+DDLOCK_WAIT,0);
      FillChar(PByte(DDSurfaceDesc.lpSurface)^,DDSurfaceDesc.lPitch*DDSurfaceDesc.dwHeight,0);
      Surface.UnLock(DDSurfaceDesc.lpSurface);
   end;
end;

Initialization

//   ResourceDLL := LoadLibrary('NeTo.Dat');

Finalization

//   FreeLibrary(ResourceDLL);

end.
