unit DxDraw;

interface

uses
   Windows, DDVer, DDraw, DDUtils, mmSystem;

const
   MAX_IMAGE_SURFACE = 2;
   MAX_AUX_SURFACE   = 1;

type

   TPixelFormat  = (pfUnknown, pfRGB8, pfRGB16, pfRGB15, pfBGR16, pfRGB24, pfBGR24, pfRGB32);

   TDirectDrawDesc = record
      IsExclusive           : Boolean;   // Direct X     
      ScreenSize            : TPoint;    // Direct Draw Screen ũ
      BPP                   : DWORD;     // Screen  Depth
      UseSystemMemory       : Boolean;   // ý ޸𸮸 ۷ 
      ImageFileName         : String[8]; // Default ̹ ̸
      MaxImage              : DWORD;     // ̹ 
      DisplayPerformance    : Boolean;
   end;

   TDirectDraw = class

      constructor Create(Handle : THandle; Desc : TDirectDrawDesc);
      destructor  Destroy; override;

   private

      FHandle           : THandle;
      FIsLocked         : Boolean;
      FErrorMessage     : String;
      FFlipPerSecond    : Double;

      procedure   DrawSurfaces;
      function    RestoreSurfaces : HResult;

      function    GetBitmapName(BitmapName : string; i : integer) : string;
      procedure   SetErrorMessage(s : String);

   public

      PixelFormat       : TPixelFormat;
      Transparency      : DWORD;
      DirectDrawDesc    : TDirectDrawDesc;

      DirectDraw        : II_DirectDraw;         // DirectDraw object  
      PrimarySurface    : II_DirectDrawSurface;  //  ȭ 
      BackBuffer        : II_DirectDrawSurface;  // ĸ 
      Clipper           : IDirectDrawClipper;   // Ŭ ü
      DDPalette         : IDirectDrawPalette;   // ĸ  Ʈʿ ε

      Image             : array[1..MAX_IMAGE_SURFACE] of II_DirectDrawSurface;
      AuxImage          : array[1..MAX_AUX_SURFACE  ] of II_DirectDrawSurface;
      LockDesc          : TT_DDSurfaceDesc;      // Lock Suface Ÿ 

      function    Initialize : Boolean;
      function    Open       : Boolean;
      function    Close      : Boolean;
      function    Finalize   : Boolean;

      procedure   FlipPage(const Boundary : TRect);

      function    MakeItSo(DDResult : HResult) : Boolean;

      procedure   LockBackGround;
      procedure   UnlockBackGround;

      property    IsLocked        : Boolean read  FIsLocked;
      property    ErrorMessage    : String  read  FErrorMessage
                                            write SetErrorMessage;

   end;


implementation

constructor TDirectDraw.Create(Handle : THandle; Desc : TDirectDrawDesc);
begin
   Inherited Create;

   FHandle        := Handle;
   DirectDrawDesc := Desc;
   Initialize;
end;

destructor TDirectDraw.Destroy;
begin
   Finalize;

   Inherited;
end;

function TDirectDraw.Initialize: Boolean;
begin
   FIsLocked  := FALSE;

   PrimarySurface := nil;
   BackBuffer     := nil;

   Initialize := TRUE;
end;

function TDirectDraw.Open: Boolean;
var
   i             : integer;
   DirectDrawOld : IDirectDraw;
   DDSurfaceDesc : TT_DDSurfaceDesc;
   ddsCaps       : TT_DDSCaps;
   DDPixelFormat : TDDPixelFormat;
   ColorKey      : TDDColorKey;
   DC            : HDC;
begin
   Open := FALSE;

   // DD COM ü 
   if DirectDrawCreate(nil, DirectDrawOld, nil) <> DD_OK then begin
      ErrorMessage := 'DirectDraw ü  ';
      Exit;
   end;

   if DirectDrawOld.QueryInterface(II_DirectDraw, DirectDraw) <> DD_OK then begin
      ErrorMessage := 'DirectDraw7 ü  ';
      Exit;
   end;

   DirectDrawOld := nil;

   if not Assigned(DirectDraw) then begin
      ErrorMessage := 'DirectDraw7 ü  ';
      Exit;
   end;

   if DirectDrawDesc.IsExclusive then begin

      // Ÿ Ǯũ Ȯ
      if DirectDraw.SetCooperativeLevel(FHandle, DDSCL_EXCLUSIVE or DDSCL_FULLSCREEN or DDSCL_ALLOWREBOOT) <> DD_OK then begin
         ErrorMessage := 'Ÿ Ǯũ  ';
         Exit;
      end;

      // ȭ  ٲ
      if DirectDraw.SetDisplayMode(DirectDrawDesc.ScreenSize.x, DirectDrawDesc.ScreenSize.y, 8 * DirectDrawDesc.BPP, 0, 0) <> DD_OK then begin
         ErrorMessage := '   ';
         Exit;
      end;

      // DirectDrawSurface ũ 
      FillChar(DDSurfaceDesc, sizeof(DDSurfaceDesc), 0);

      DDSurfaceDesc.dwSize            := sizeof(DDSurfaceDesc);
      DDSurfaceDesc.dwFlags           := DDSD_CAPS or DDSD_BACKBUFFERCOUNT;
      DDSurfaceDesc.ddsCaps.dwCaps    := DDSCAPS_COMPLEX or DDSCAPS_FLIP or DDSCAPS_PRIMARYSURFACE;
      if DirectDrawDesc.UseSystemMemory then
         DDSurfaceDesc.ddsCaps.dwCaps := DDSurfaceDesc.ddsCaps.dwCaps or DDSCAPS_SYSTEMMEMORY;
      DDSurfaceDesc.dwBackBufferCount := 1;

      //  ȭ  
      if DirectDraw.CreateSurface(DDSurfaceDesc, PrimarySurface, nil) <> DD_OK then begin
         ErrorMessage := ' ȭ   ';
         Exit;
      end;

      // ĸ  ü 
      FillChar(ddsCaps, sizeof(ddsCaps), 0);
      ddsCaps.dwCaps := DDSCAPS_BACKBUFFER;
      if PrimarySurface.GetAttachedSurface(ddsCaps, BackBuffer) <> DD_OK then begin
         ErrorMessage := 'ĸ   ';
         Exit;
      end;

   end else begin

      // Ϲ  
      if DirectDraw.SetCooperativeLevel(FHandle, DDSCL_NORMAL) <> DD_OK then begin
         ErrorMessage := 'Ϲ    ';
         Exit;
      end;

      // DirectDrawSurface ũ 
      FillChar(DDSurfaceDesc, sizeof(DDSurfaceDesc), 0);
      with DDSurfaceDesc do begin
         dwSize            := sizeof(DDSurfaceDesc);
         dwFlags           := DDSD_CAPS;
         ddsCaps.dwCaps    := DDSCAPS_PRIMARYSURFACE;
      end;

      //  ȭ  
      if DirectDraw.CreateSurface(DDSurfaceDesc, PrimarySurface, nil) <> DD_OK then begin
         ErrorMessage := ' ȭ   ';
         Exit;
      end;

      // ĸ  ü 
      FillChar(DDSurfaceDesc, sizeof(DDSurfaceDesc), 0);

      DDSurfaceDesc.dwSize            := sizeof(DDSurfaceDesc);
      DDSurfaceDesc.dwFlags           := DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH;
      DDSurfaceDesc.ddsCaps.dwCaps    := DDSCAPS_OFFSCREENPLAIN;
      if DirectDrawDesc.UseSystemMemory then
         DDSurfaceDesc.ddsCaps.dwCaps := DDSurfaceDesc.ddsCaps.dwCaps or DDSCAPS_SYSTEMMEMORY;
      DDSurfaceDesc.dwWidth           := DirectDrawDesc.ScreenSize.x;
      DDSurfaceDesc.dwHeight          := DirectDrawDesc.ScreenSize.y;

      if DirectDraw.CreateSurface(DDSurfaceDesc, BackBuffer, nil) <> DD_OK then begin
         ErrorMessage := '   ';
         Exit;
      end;

      if DirectDraw.CreateClipper(0, Clipper, nil) <> DD_OK then begin
         ErrorMessage := 'Ŭ ü  ';
         Exit;
      end;

      if Clipper.SetHWnd(0, FHandle) <> DD_OK then begin
         ErrorMessage := 'Ŭ ü  ';
         Exit;
      end;

      if PrimarySurface.SetClipper(Clipper) <> DD_OK then begin
         ErrorMessage := 'Ŭ ǥ ߰ ';
         Exit;
      end;

      try
         BackBuffer.GetDC(DC);
         DirectDrawDesc.BPP := GetDeviceCaps(DC, BITSPIXEL) div 8;
      finally
         BackBuffer.ReleaseDC(DC);
      end;
   end;

   // ȷƮ ε
   if (DirectDrawDesc.BPP = 1) and DirectDrawDesc.IsExclusive then begin
      DDPalette := DDLoadPalette(DirectDraw, GetBitmapName(DirectDrawDesc.ImageFileName, 1));
      if PrimarySurface.SetPalette(DDPalette) <> DD_OK then begin
         ErrorMessage := 'ȷƮ ε ';
         Exit;
      end;
   end;

   // Ʈ  & ε
   for i := 1 to DirectDrawDesc.MaxImage do begin
      Image[i] := DDLoadBitmap(DirectDraw, GetBitmapName(DirectDrawDesc.ImageFileName, i), 0, 0, DirectDrawDesc.UseSystemMemory);
      //  
      ColorKey.dwColorSpaceLowValue  := 0;
      ColorKey.dwColorSpaceHighValue := 0;
      if Image[i].SetColorKey(DDCKEY_SRCBLT, @ColorKey) <> DD_OK then begin
         ErrorMessage := '  ';
         Exit;
      end;
   end;

   // ȼ  ǵ
   FillChar(DDPixelFormat, sizeof(DDPixelFormat), 0);
   DDPixelFormat.dwSize := sizeof(DDPixelFormat);
   if PrimarySurface.GetPixelFormat(DDPixelFormat) <> DD_OK then begin
      ErrorMessage := 'ȼ  б ';
      Exit;
   end;

   case DirectDrawDesc.BPP of
      1 :
      begin
         PixelFormat := pfRGB8;
      end;

      2 :
      begin
         if DDPixelFormat.dwRBitMask = $0000F800 then begin
            PixelFormat   := pfRGB16;
            Transparency  := $FFFF;
         end else if DDPixelFormat.dwRBitMask = $00007C00 then begin
            PixelFormat   := pfRGB15;
            Transparency  := $7FFF;
         end else if DDPixelFormat.dwRBitMask = $0000001F then begin
            PixelFormat   := pfBGR16;
            Transparency  := $FFFF;
         end else begin
            PixelFormat   := pfUnknown;
            Transparency  := $0000;
         end;
      end;

      3 :
      begin
         if DDPixelFormat.dwRBitMask = $00FF0000 then begin
            PixelFormat   := pfRGB24;
            Transparency := $FFFFFF;
         end else if DDPixelFormat.dwRBitMask = $000000FF then begin
            PixelFormat   := pfBGR24;
            Transparency := $FFFFFF;
         end else begin
            PixelFormat   := pfUnknown;
            Transparency := $000000;
         end;
      end;

      4 :
      begin
         PixelFormat   := pfRGB32;
         Transparency := $FFFFFF;
      end;

      else
      begin
         PixelFormat   := pfUnknown;
      end;

   end;

   if DirectDrawDesc.IsExclusive then begin
      ShowCursor(FALSE);
   end;

   Open := TRUE;
end;

function TDirectDraw.Close: Boolean;
var
   i : Integer;
begin
   UnlockBackGround;
   
   // ȭ  
   if Assigned(PrimarySurface) then
      PrimarySurface := nil;

   if DirectDrawDesc.IsExclusive then begin
      Pointer(BackBuffer) := nil;
   end
   else begin
      if Assigned(BackBuffer) then
         BackBuffer := nil;
   end;

   if Assigned(Clipper) then
      Clipper := nil;

   if Assigned(DDPalette) then
      DDPalette := nil;

   for i := 1 to MAX_IMAGE_SURFACE do begin
      if Assigned(Image[i]) then
         Image[i] := nil;
   end;

   for i := 1 to MAX_AUX_SURFACE do begin
      if Assigned(AuxImage[i]) then
         AuxImage[i] := nil;
   end;

   // DirectDraw COM ü 
   if Assigned(DirectDraw) then
      DirectDraw := nil;

   ShowCursor(TRUE);

   Close := TRUE;
end;

function TDirectDraw.Finalize: Boolean;
begin
   Finalize := TRUE;
end;

procedure TDirectDraw.DrawSurfaces;
var
   i : integer;
begin
   for i := 1 to DirectDrawDesc.MaxImage do begin
//      DDReloadBitmap(Image[i], GetBitmapName(DirectDrawDesc.ImageFileName,i));
   end;
end;

function TDirectDraw.MakeItSo(DDResult: HResult): Boolean;
begin
   // ȭ   Լ
   case DDResult of
      DD_OK             : Result := TRUE;
      DDERR_SURFACELOST : Result := RestoreSurfaces <> DD_OK;
      else                Result := DDResult <> DDERR_WASSTILLDRAWING;
   end;
end;

function TDirectDraw.RestoreSurfaces: HResult;
var
   i : integer;
begin
   // ȭ Ҿ MakeItSo ȣ
   Result := PrimarySurface._Restore;
   if Result = DD_OK then begin
      for i := 1 to DirectDrawDesc.MaxImage do begin
         Result := Image[i]._Restore;
         if Result <> DD_OK then
            Exit;
     end;
      if Result = DD_OK then
         DrawSurfaces;
   end;
end;

function TDirectDraw.GetBitmapName(BitmapName: string; i: integer): string;
var
   s : String;
begin
   Str(i, s);
   Result := BitmapName + s;
end;

procedure TDirectDraw.SetErrorMessage(s: String);
begin
   FErrorMessage := s;
end;

procedure TDirectDraw.LockBackGround;
begin
   if not FIsLocked then begin
      LockDesc.dwSize := sizeof(LockDesc);
      repeat
      until MakeItSo(BackBuffer.Lock(PRect(nil),LockDesc,DDLOCK_SURFACEMEMORYPTR + DDLOCK_WAIT, 0));
      FIsLocked := TRUE;
   end;
end;

procedure TDirectDraw.UnlockBackGround;
begin
   if FIsLocked then begin
      BackBuffer.UnLock(LockDesc.lpSurface);
      FIsLocked := FALSE;
   end;
end;

procedure TDirectDraw.FlipPage(const Boundary : TRect);
const
   IsFirst   : Boolean = TRUE;
   StartTime : DWORD = 0;
   FlipCount : Integer = 0;
var
   DC        : HDC;
   FPS       : String;
   TimeGap   : Longint;
begin
   Inc(FlipCount);

   if DirectDrawDesc.DisplayPerformance then begin
      if IsFirst then begin
         StartTime := timeGetTime;
         IsFirst   := FALSE;
      end
      else begin
         TimeGap        := (timeGetTime - StartTime);
         FFlipPerSecond := (FlipCount * 1000) / TimeGap;
         Str(FFlipPerSecond : 3:2, FPS);

         BackBuffer.GetDC(DC);
         TextOut(DC, 0, 0, PChar(FPS), Length(FPS));
         BackBuffer.ReleaseDC(DC);
      end;
   end;

   if DirectDrawDesc.IsExclusive then begin
      repeat
      until MakeItSo(PrimarySurface.Flip(nil, DDFLIP_WAIT));
   end else begin
      DirectDraw.WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN, 0);
      repeat
      until MakeItSo(PrimarySurface.Blt(@Boundary, BackBuffer, PRect(nil), DDFLIP_WAIT, PDDBltFX(nil)));
   end;
end;

end.
