unit LiteMain;

interface

uses
   Windows, DxDraw, DDUtils, LiteUtil;

const
   WIN_MOUSE_X_GAP  = 4;
   WIN_TITLE_HEIGHT = 20;
   WIN_MOUSE_Y_GAP  = WIN_TITLE_HEIGHT + WIN_MOUSE_X_GAP;

type
   TLiteMain = class
	  constructor Create(Handle : HWND);
	  destructor  Destroy; override;

   private
	  ParentHandle  : HWND;
	  SystemDesc    : TDirectDrawDesc;
	  LoopCount     : Integer;

   public
	  DirectDraw    : TDirectDraw;

	  procedure   GameLoop;
   end;

implementation

{ TLiteMain }

constructor TLiteMain.Create(Handle : HWND);
begin
   inherited Create;

   LoopCount     := 0;

   ParentHandle  := Handle;

   // Direct Draw κ
   SystemDesc.IsExclusive        := TRUE;
   SystemDesc.ScreenSize.x       := 640;
   SystemDesc.ScreenSize.y       := 480;
   SystemDesc.BPP                := 1;
   SystemDesc.UseSystemMemory    := not TRUE;
   SystemDesc.ImageFileName      := 'Neto';
   SystemDesc.MaxImage           := 2;
   SystemDesc.DisplayPerformance := TRUE;

   DirectDraw := TDirectDraw.Create(Handle, SystemDesc);
   if not DirectDraw.Open then
	  PostQuitMessage(0);

end;

destructor TLiteMain.Destroy;
begin
   DirectDraw.Close;
   DirectDraw.Free;

   inherited;
end;

procedure TLiteMain.GameLoop;
var
   Boundary    : TRect;
   LocalCount  : Integer;
   Index, x, y : Integer;
begin
   Inc(LoopCount);

   DDClearDevice(DirectDraw.BackBuffer, 255);

   LocalCount := (LoopCount div 10) mod 16;

   Index    := LocalCount div 8;
   x        := (LocalCount mod 8) mod 4;
   y        := (LocalCount mod 8) div 4;
   Boundary := Bounds(x * (640 div 4), y * (480 div 2), 640 div 4, 480 div 2);

   for y := 0 to Pred(2) do
   for x := 0 to Pred(4) do begin
      DirectDraw.BackBuffer.BltFast(x * (640 div 4), y * (480 div 2), DirectDraw.Image[Index+1], @Boundary, 0);
   end;

   if SystemDesc.IsExclusive then begin
      DirectDraw.FlipPage(PRect(nil)^);
   end
   else begin
      GetWindowRect(ParentHandle, Boundary);
      Boundary := Bounds(Boundary.Left + WIN_MOUSE_X_GAP, Boundary.Top + WIN_MOUSE_Y_GAP,
                         SystemDesc.ScreenSize.x, SystemDesc.ScreenSize.y);
      DirectDraw.FlipPage(Boundary);
   end;
end;

end.

